###############################################################################
# Copyright (c) 2014 BC Cancer Agency, NIH
# 
# License
# The software is distributed under the terms of the 
# GNU Lesser General Public License (LGPL)
# 
# Disclaimer
# This software and documentation come with no warranties of any kind.
# This software is provided "as is" and any and any express or implied 
# warranties, including, but not limited to, the implied warranties of
# merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the  copyright holder be liable for any direct, 
# indirect, incidental, special, exemplary, or consequential damages
# (including but not limited to, procurement of substitute goods or 
# services; loss of use, data or profits; or business interruption)
# however caused and on any theory of liability, whether in contract,
# strict liability, or tort arising in any way out of the use of this 
# software.
# 
###############################################################################

applyGatingML2 <- function(...) {
  options("warn"=-1)
  suppressPackageStartupMessages(suppressMessages(.applyGatingML2(...)))
}

.applyGatingML2 <- function(...) {
  args <- c(...)

  # Parse command line parameters
  fcsFile <- NULL     # input FCS data file
  gateFile <- NULL    # input Gating-ML file
  outputBase <- NULL  # for the output file names, csv and summary, (optional, default fcsFile) 
  figures <- NULL     # png | none (default)

  vers <- "3.0"
  offlinetest = FALSE  # This will be rewritten if -t is detected, which means we are running without GenePattern server
  libdir <- ''
  server.dir <- ''
  patch.dir <- ''

  for(i in 1:length(args)) {
    flag <- substring(args[[i]], 0, 4)
    arg <- substring(args[[i]], 5, nchar(args[[i]]))
  
    if(flag == '-ii=') {
      fcsFile <- arg     
    } else if (flag == '-ig=') {
      gateFile <- arg
    } else if (flag == '-bi=') {
      outputBase <- arg
    } else if (flag == '-fi=') {
      if(arg == 'png') { figures <- arg } else { figures <- NULL }
    } else if (flag == '-li=') {
        libdir <- arg
    } else if (flag == '-si=') {
        server.dir <- arg
    } else if (flag == '-pi=') {
        patch.dir <- arg
    } else if (flag == '-t') {
      offlinetest = TRUE
    } else {
      stop(paste("Unknown option", flag, sep=": "), call.=FALSE)
    }
  }

  source(paste(libdir, "common.R", sep=''))
  source(paste(libdir, "loadRLibrary.R", sep=''))
  source(paste(libdir, "helperFunc.R", sep=''))
  if (offlinetest) {
    load.packages.offline(libdir, verbose=FALSE)
  } else {
    load.packages(libdir, patch.dir, server.dir, vers, verbose=FALSE)  
  }
  
  # Validate parameters
  if(is.null(fcsFile)) {
    stop("You must specify a data file!", call.=FALSE)
  }
  if(is.null(gateFile)) {
    stop("You must specify a Gating-ML file", call.=FALSE)
  }
  if(is.null(outputBase)) {
    outputBase = fcsFile;
  }

  flowEnv=new.env()
  read.gatingML(gateFile, flowEnv)
  if (flowEnv$GatingMLVersion >= 2) {
    fcs <- read.FCS(fcsFile, transformation="linearize-with-PnG-scaling") 
  }
  else {
    fcs <- read.FCS(fcsFile, transformation=FALSE)  
  }
  
  g <- ls(flowEnv)
  gates <- c()
  for(gg in g) {
    # MYidnum always appears in the environment and it is not a gate
    # There are also some functions in the environment
    if((gg != 'MYidnum') && (gg != 'GatingMLVersion') && (mode(get(gg,envir=flowEnv)) == 'S4') && (class(get(gg,envir=flowEnv)) != 'compensation')) { gates <- c(gates, gg) } 
  }

  for(gateId in gates) {
    gate<-flowEnv[[gateId]]
    result<-filter(fcs,gate)
    x<-split(fcs,result)[1]
    frame<-x[[names(x)]]
    try(writeCSVFile(frame, paste(outputBase, gateId, 'csv', sep=".")))
    if(result@filterId == '') { result@filterId <- gateId }
    try(writeGateSummary(frame, fcs, gateId, paste(outputBase, gateId, 'summary', 'txt', sep=".")))
    if((!is.null(figures)) && (length(frame@exprs) > 0)) {
      try(plotFile(frame, paste(outputBase, gateId, 'figure', 'png', sep=".")))
    }
  }

}
 
plotFile <- function (frame, filename, ...) {
  png(file=filename, ...)  
  plot(frame)
  dev.off()
}

writeGateSummary <- function(frame, fcs, gateId, filename) {
  sink(file=filename)
  eventsInFCS <- length(fcs@exprs)/as.integer(fcs@description[['$PAR']])
  eventsInGate <- length(frame@exprs)/as.integer(frame@description[['$PAR']])
  eventsPercentage <- 100 * eventsInGate / eventsInFCS
  cat(paste(eventsInGate, ' out of ', eventsInFCS, ' in gate ', gateId, ' (', round(eventsPercentage, 2), ' %).\n\n', sep=''))
  print(summary(frame))
  sink()
}
 
writeCSVFile <- function(frame, filename) {
  sink(file=filename)  
  cat( sapply(1:ncol(frame), function (x) getNameAndDescription(frame, x)), sep=",")  
  cat("\n")
  apply(exprs(frame), 1, writeCSVDataLine)
  sink()
}
 
getNameAndDescription <- function(frame, i) {
  shortName <- NULL
  name <- NULL
  shortName <- frame@description[[paste("$P", i,"N", sep="")]]
  name <- frame@description[[paste("$P", i,"S", sep="")]]
  if(!is.null(shortName) && !is.null(name)) {
    return( paste(shortName, name, sep=":"))
  } else if(!is.null(shortName)) {
    return(shortName)
  } else if(!is.null(name)) {
    return(name)
  } else {
    return(paste("$P", i, "N", sep=""))
  }      
}

writeCSVDataLine <- function (x) {
  cat(x, sep=",")
  cat("\n")
}

