###############################################################################
# Copyright (c) 2014 Josef Spidlen, BC Cancer Agency
# 
# License
# The software is distributed under the terms of the 
# GNU Lesser General Public License (LGPL)
# 
# Disclaimer
# This software and documentation come with no warranties of any kind.
# This software is provided "as is" and any and any express or implied 
# warranties, including, but not limited to, the implied warranties of
# merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the  copyright holder be liable for any direct, 
# indirect, incidental, special, exemplary, or consequential damages
# (including but not limited to, procurement of substitute goods or 
# services; loss of use, data or profits; or business interruption)
# however caused and on any theory of liability, whether in contract,
# strict liability, or tort arising in any way out of the use of this 
# software.
# 
###############################################################################


## Get fluorescence channel name, i.e., channels that are not time, forward or side scatter
## existingChannels is a vector of channel names
## return value a vector of channel names that is a subset of the input set
getFluorescenceChannelNames <- function(existingChannels) {
  channel.names <- NULL;
  for (name in existingChannels) {
    if (
      (nchar(name) - nchar(gsub('time', '', name, ignore.case = TRUE)) == 0) &&
        (nchar(name) - nchar(gsub('fsc', '', name, ignore.case = TRUE)) == 0) &&
        (nchar(name) - nchar(gsub('ssc', '', name, ignore.case = TRUE)) == 0)
    ) 
      channel.names <- c(channel.names, name)
  }
  return(channel.names)
}

## Parse the string channels and return a vector of channel names from the channels parameter 
## that exist in the existingChannels parameter.
## Channels is a comma separated list of either channel indexes or channel names, 
## i.e., "1,2,3,5" or "FL1-H, FL2-A, FL3-H". 
## White spaces are OK (i.e., ignored)
getSpecifiedChannelNames <- function(channels, existingChannels) {
  channel.names <- unlist(strsplit(channels,","))
  # Remove whitespaces 
  for(i in 1:length(channel.names)) {
    while((nchar(channel.names[i]) > 0) && substring(channel.names[i], 0, 1) == ' ') channel.names[i] <- substring(channel.names[i], 2, nchar(channel.names[i]))
    while((nchar(channel.names[i]) > 0) && substring(channel.names[i], nchar(channel.names[i]), nchar(channel.names[i])) == ' ') channel.names[i] <- substring(channel.names[i], 1, nchar(channel.names[i]) - 1)
  }
  
  if((substring(channel.names[1], 0, 1) >= '1') && (substring(channel.names[1], 0, 1) <= '9')) {
    # They used indexes rather than names, need to convert these...
    channel.indexes <- channel.names
    channel.names <- NULL
    for(i in 1:length(channel.indexes)) {
      if(as.integer(channel.indexes[i]) <= length(existingChannels)) {
        channel.names <- c(channel.names, existingChannels[as.integer(channel.indexes[i])])
      } else {
        print(paste("Warning, channel", channel.indexes[i], "will be skipped."))
      }
    }
  } else {
    # Verify that all specified channel names can be found in existing channels
    channel.names <- verifyChannelNames(channel.names, existingChannels)
  } 
  
  return(channel.names) 
}

## Check that channelSubset is contained in channelSet. Return only the channels that are there.
verifyChannelNames <- function(channelSubset, channelSet) {
  tmp <- NULL
  for(name in channelSubset) {
    if(name %in% channelSet) {
      tmp <- c(tmp, name)
    } else {
      print(paste("Warning, channel", name, "not found in the source data, it will be skipped."))
    }
  }
  return(tmp)
}

## Loading required libraries (for testing purposes without using GenePattern)
load.packages.offline <- function(libdir, verbose=FALSE) {
  r.packages <- read.table(paste(libdir, "r.libraries.txt", sep=""))
  r.packages <- unlist(r.packages)
  for(i in 1:length(r.packages)) {
    r.lib.spec <- toString(r.packages[[i]])
    r.lib.name    <- strsplit(r.lib.spec, "_")[[1]][[1]]
    r.lib.version <- strsplit(r.lib.spec, "_")[[1]][[2]]
    if (verbose) {
      cat("\nLoading library: ") 
      cat(r.lib.name)
    }
    suppressPackageStartupMessages(library(r.lib.name, character.only=TRUE))
  }
  if (verbose) cat("\nDone loading R libraries.\n")
}
